"""
sphinx.ext.autodoc extension for classes with configurable traits.

The code here is similar to the official code example in
https://www.sphinx-doc.org/en/master/development/tutorials/autodoc_ext.html#writing-the-extension.
"""

import warnings

from sphinx.ext.autodoc import AttributeDocumenter, ClassDocumenter, ObjectMember
from sphinx.util.inspect import safe_getattr
from traitlets import MetaHasTraits, TraitType, Undefined

# __version__ should be updated using tbump, based on configuration in
# pyproject.toml, according to instructions in RELEASE.md.
#
__version__ = "1.2.2"


class ConfigurableDocumenter(ClassDocumenter):
    """
    Specialized Documenter subclass for traits with config=True

    Links to relevant source code in sphinx.ext.autodoc:
    - Documenter:      https://github.com/sphinx-doc/sphinx/blob/v6.0.0b2/sphinx/ext/autodoc/__init__.py#L270-L299
    - ClassDocumenter: https://github.com/sphinx-doc/sphinx/blob/v6.0.0b2/sphinx/ext/autodoc/__init__.py#L1395-L1408
    """

    # objtype: The suffix to "auto" for a Sphinx directive name that will be
    #          created (and the default value for "directivetype").
    objtype = "configurable"

    # directivetype: Clarifies that this Documenter class could be capable of
    #                documenting this kind of members of other parent like
    #                Documenter classes.
    directivetype = "class"

    # priority: Declares this class' priority for use if multiple classes
    #           "can_document_member" of the "directivetype". This is only
    #           relevant if a parent class has a member.
    #
    #           ConfigurableDocumenter can document traitlets configurable
    #           classes, so a parent like Documenter class can be the
    #           ModuleDocumenter.
    #
    priority = 100  # higher priority than ClassDocumenter's 10

    @classmethod
    def can_document_member(cls, member, membername, isattr, parent):
        """
        If the member is a class with traitlets then we can document it, and we
        will document it thanks to a high priority.

        This function is not considered if the ``autoconfigurable`` or
        ``autoclass`` directives are called directly. can_document_member is
        only used by other parent like Documenter classes having members of this
        class' configured "documentertype" - such as ModuleDocumenter.
        """
        return isinstance(member, MetaHasTraits)

    def get_object_members(self, want_all):
        """
        This get_object_members function override is a hack, manipulating
        __doc__ values of trait configuration objects, but otherwise behaving
        exactly like the super class get_object_members.

        It sets truthy strings to the class' traits __doc__ attributes. They
        will otherwise be filtered out by the Documenter.filter_members
        function, unless ``undoc-members`` option is set.

        Links to relevant source code in sphinx.ext.autodoc:
        - Documenter.filter_members:          https://github.com/sphinx-doc/sphinx/blob/v6.0.0b2/sphinx/ext/autodoc/__init__.py#L616-L769
        - Documenter.get_object_members:      https://github.com/sphinx-doc/sphinx/blob/v6.0.0b2/sphinx/ext/autodoc/__init__.py#L607-L614
        - ClassDocumenter.get_object_members: https://github.com/sphinx-doc/sphinx/blob/v6.0.0b2/sphinx/ext/autodoc/__init__.py#L1656-L1674
        - traitlets.HasTraits.class_traits:   https://github.com/ipython/traitlets/blob/v5.6.0/traitlets/traitlets.py#L1620-L1652
        """
        check, members = super().get_object_members(want_all)
        existing_member_names = {m.__name__ for m in members}

        # We add all _configurable_ traits, including inherited, even if they
        # weren't included via :members: or :inherited-members: options.
        #
        # Being added at this stage doesn't mean they are presented. Any members
        # added here must also have a truthy __doc__ string attribute and not be
        # part of the :exclude-members: option.
        #
        # FIXME: We have been adding the configurable trait_members
        #        unconditionally, but should we keep doing that?
        #
        #        See https://github.com/jupyterhub/autodoc-traits/issues/27
        #
        config_trait_members = self.object.class_traits(config=True).items()
        for trait_tuple in config_trait_members:
            name, trait = trait_tuple
            if not trait.__doc__:
                warnings.warn(
                    f"""
                    Documenting {self.object.__name__}.{trait.name} without a help string because it has config=True.

                    Including undocumented config=True traits is deprecated in autodoc-traits 1.1.
                    Add a help string:

                        {trait.name} = {trait.__class__.__name__}(
                            help="...",
                        )

                    to keep this trait in your docs,
                    or include it explicitly via :members:
                    """,
                    FutureWarning,
                )
                # FIXME: in the unlikely event that the patched trait
                # is documented multiple times in the same build,
                # this patch will cause it to have a truthy help string
                # elsewhere, not just in this autoconfigurable instance.
                trait.__doc__ = trait.help = "No help string is provided."
            if name not in existing_member_names:
                existing_member_names.add(name)

                members.append(ObjectMember(name, trait))

        return check, members


class TraitDocumenter(AttributeDocumenter):
    """
    Links to relevant source code in sphinx.ext.autodoc:
    - Documenter:          https://github.com/sphinx-doc/sphinx/blob/v6.0.0b2/sphinx/ext/autodoc/__init__.py#L270-L299
    - AttributeDocumenter: https://github.com/sphinx-doc/sphinx/blob/v6.0.0b2/sphinx/ext/autodoc/__init__.py#L2509-L2524
    """

    # objtype: The suffix to "auto" for a Sphinx directive name that will be
    #          created (and the default value for "directivetype").
    objtype = "trait"

    # directivetype: Clarifies that this Documenter class could be capable of
    #                documenting this kind of members of other parent like
    #                Documenter classes.
    directivetype = "attribute"

    # priority: Declares this class' priority for use if multiple classes
    #           "can_document_member" of the "directivetype". This is only
    #           relevant if a parent class has a member.
    #
    #           TraitsDocumenter can document traitlets type attributes, so the
    #           parent like Documenter class is typically ClassDocumenter, but
    #           can also be the ConfigurableDocumenter.
    #
    priority = 100  # AttributeDocumenter has 10

    # order: order if the autodoc_member_order in conf.py is set to "groupwise",
    #        by default it is "alphabetical", where lowest order comes first.
    #        Since traits are relevant configuration, we declare the lowest
    #        order for high visual priority.
    member_order = 0  # AttributeDocumenter has 60

    @classmethod
    def can_document_member(cls, member, membername, isattr, parent):
        """
        If the member is a traitlets type we can document it, and we will
        document it thanks to a high priority.

        This function is not considered if the ``autotrait`` or
        ``autoattribute`` directives are called directly. can_document_member is
        only used by other parent like Documenter classes having members of this
        class' configured "documentertype" - such as ClassDocumenter.
        """
        return isinstance(member, TraitType)

    def add_directive_header(self, sig):
        """
        add_directive_header is called by the base class' Documenter.generate
        method. It is provided by both AttributeDocumenter and Documenter. This
        override retains use of the super classes implementations, but influence
        them.

        For functions, the directive header describes the function's call
        signature, but not the function's docstring.

        Similarly, we look to emit rST to describe how the traitlets
        configuration option can be configured and its default value.

        - Documenter.generate:                      https://github.com/sphinx-doc/sphinx/blob/v6.0.0b2/sphinx/ext/autodoc/__init__.py#L918-L929
        - AttributeDocumenter.add_directive_header: https://github.com/sphinx-doc/sphinx/blob/v6.0.0b2/sphinx/ext/autodoc/__init__.py#L2592-L2620
        - Documenter.add_directive_header:          https://github.com/sphinx-doc/sphinx/blob/v6.0.0b2/sphinx/ext/autodoc/__init__.py#L504-L524
        """
        default_value = self.object.default_value
        if default_value is Undefined:
            default_value = ""
        else:
            default_value = repr(default_value)

        traitlets_type = self.object.__class__.__name__  # Bool

        if self.object.metadata.get("config"):
            # add config prefix (c.TestConfigurator.trait = ) if it's configurable
            config_prefix = f"c.{self.format_name()} = "
        else:
            config_prefix = ""

        self.options.annotation = f"{config_prefix}{traitlets_type}({default_value})"

        super().add_directive_header(sig)


def hastraits_attrgetter(obj, name, *defargs):
    """getattr for trait

    Ensures when HasTraits are documented, their __doc__ attr is defined
    as the .help string.

    This backports a change in traitlets 5.8.0.
    """
    attr = safe_getattr(obj, name, *defargs)
    if isinstance(attr, TraitType):
        # ensure __doc__ is defined as the trait's help string
        # if help is empty, that's the same as undocumented
        attr.__doc__ = attr.help
    return attr


def setup(app):
    """
    The setup function is required for Sphinx extensions.

    In this function we register the sphinx.ext.autodoc extension that this
    extension needs to function, and we register our sphinx.ext.autodoc
    Documenter classes we provide.
    """
    # setup_extension reference:
    # https://www.sphinx-doc.org/en/master/extdev/appapi.html#sphinx.application.Sphinx.setup_extension
    app.setup_extension("sphinx.ext.autodoc")

    # add_autodocumenter reference:
    # https://www.sphinx-doc.org/en/master/extdev/appapi.html#sphinx.application.Sphinx.add_autodocumenter
    app.add_autodocumenter(ConfigurableDocumenter)
    app.add_autodocumenter(TraitDocumenter)

    # add_autodoc_attrgetter reference:
    # https://www.sphinx-doc.org/en/master/extdev/appapi.html#sphinx.application.Sphinx.add_autodoc_attrgetter
    app.add_autodoc_attrgetter(MetaHasTraits, hastraits_attrgetter)



import glob
import os
import shutil
import sys
import tempfile
from functools import partial

import pytest

if sys.version_info >= (3, 8):
    copy_tree = partial(shutil.copytree, dirs_exist_ok=True)
else:
    # use deprecated distutils on Python < 3.8
    # when shutil.copytree added dirs_exist_ok support
    from distutils.dir_util import copy_tree


@pytest.fixture
def temp_docs_dir():
    """
    This fixture provides a temporary directory with files copied from the
    tests/docs directory.
    """
    with tempfile.TemporaryDirectory() as temp_dir:
        tests_dir = os.path.abspath(os.path.dirname(__file__))
        test_docs_dir = os.path.join(tests_dir, "docs")

        # populate content
        copy_tree(test_docs_dir, temp_dir)

        yield temp_dir


@pytest.fixture
def get_glob_filtered_temp_docs_dir(temp_docs_dir):
    """
    This fixture provides function that returns a path to a temp docs directory
    based on tests/docs, but filtered to only retain .rst files globbed by
    provide glob_patterns relative to the source/ directory.

    Note that to test specific documents, including those that are expected to
    raise errors, we could use the Sphinx configuration "include_patterns".
    Sadly its only available in Sphinx 5.1+ so it would constrain us to test
    against Sphinx 5.1+. Due to that, we rely on this fixture instead.
    """

    def _filter_source_dir_func(glob_patterns):
        old_cwd = os.getcwd()
        try:
            os.chdir(os.path.join(temp_docs_dir, "source"))
            source_rst_files = set(glob.glob("**/*.rst", recursive=True))
            files_to_retain = set()
            for p in glob_patterns:
                files_to_retain = files_to_retain.union(
                    set(glob.glob(p, recursive=True))
                )

            if not source_rst_files.intersection(files_to_retain):
                print("glob_patterns", glob_patterns)
                print("source_rst_files", source_rst_files)
                print("files_to_retain", files_to_retain)
                raise ValueError(
                    "provided glob_patterns found no .rst in the source folder to retain!"
                )

            for f in source_rst_files.difference(files_to_retain):
                os.remove(f)

            print()
            print(
                f"Fixture get_glob_filtered_temp_docs_dir provided the directory {temp_docs_dir}:"
            )
            for f in glob.glob("**/*.rst", recursive=True):
                print(f"- {f}")
            print()
        finally:
            os.chdir(old_cwd)
        return temp_docs_dir

    yield _filter_source_dir_func



"""
sample_module docstring

This module provides a module and classes with and without traits to test
autodoc_traits against. It does not contain tests.
"""

from traitlets import Bool, Dict, Instance, Integer, Unicode, Union
from traitlets.config.configurable import Configurable


class SampleConfigurable(Configurable):
    """SampleConfigurable docstring"""

    non_trait = False

    @property
    def non_trait_property(self):
        """non_trait_property docstring"""

    trait = Bool(
        help="""trait help text""",
        config=True,
    )
    trait_nohelp = Bool(
        config=True,
    )
    trait_noconfig = Bool(
        help="""trait_noconfig help text""",
    )

    def method(self):
        """method docstring"""


class SampleConfigurableSubclass(SampleConfigurable):
    """SampleConfigurableSubclass docstring"""

    subclass_non_trait = False

    @property
    def subclass_non_trait_property(self):
        """subclass_non_trait_property docstring"""

    subclass_trait = Bool(
        config=True,
        help="""subclass_trait help text""",
    )
    subclass_trait_nohelp = Bool(
        config=True,
    )
    subclass_trait_noconfig = Bool(
        help="""subclass_trait_noconfig help text""",
    )

    def subclass_method(self):
        """subclass_method docstring"""


class SampleNonConfigurable:
    """SampleNonConfigurable docstring"""

    non_trait = False

    @property
    def non_trait_property(self):
        """non_trait_property docstring"""

    def method(self):
        """method docstring"""


class SampleNonConfigurableSubclass(SampleNonConfigurable):
    """SampleNonConfigurableSubclass docstring"""

    non_trait = False

    @property
    def subclass_non_trait_property(self):
        """subclass_non_trait_property docstring"""

    def subclass_method(self):
        """subclass_method docstring"""


class TraitTypesSampleConfigurable(Configurable):
    """TraitTypesSampleConfigurable docstring"""

    trait_integer = Integer(
        help="""trait_integer help text""",
        config=True,
    )
    trait_integer_nohelp = Integer(
        config=True,
    )
    trait_integer_noconfig = Integer(
        help="""trait_integer_noconfig help text""",
    )

    trait_unicode = Unicode(
        help="""trait_unicode help text""",
        config=True,
    )
    trait_unicode_nohelp = Unicode(
        config=True,
    )
    trait_unicode_noconfig = Unicode(
        help="""trait_unicode_noconfig help text""",
    )

    trait_dict = Dict(
        help="""trait_dict help text""",
        config=True,
    )
    trait_dict_nohelp = Dict(
        config=True,
    )
    trait_dict_noconfig = Dict(
        help="""trait_dict_noconfig help text""",
    )

    trait_instance = Instance(
        klass=SampleConfigurable,
        help="""trait_instance help text""",
        config=True,
    )
    trait_instance_nohelp = Instance(
        klass=SampleConfigurable,
        config=True,
    )
    trait_instance_noconfig = Instance(
        klass=SampleConfigurable,
        help="""trait_instance_noconfig help text""",
    )

    trait_union = Union(
        [Integer(), Unicode()],
        help="""trait_union help text""",
        config=True,
    )
    trait_union_nohelp = Union(
        [Integer(), Unicode()],
        config=True,
    )
    trait_union_noconfig = Union(
        [Integer(), Unicode()],
        help="""trait_union_noconfig help text""",
    )



import os
import subprocess

import pytest


def test_sphinx_build_all_docs(temp_docs_dir, monkeypatch):
    """
    Tests that the docs folder builds without warnings.
    """
    monkeypatch.chdir(temp_docs_dir)

    subprocess.run(
        ["sphinx-build", "--color", "-W", "--keep-going", "source", "build"],
        check=True,
        text=True,
    )


@pytest.mark.parametrize(
    "rst_file_to_test, strings_in_html, strings_not_in_html",
    [
        (
            "autoclass/members.rst",
            [
                "c.SampleConfigurable.trait",
                "trait_noconfig",
            ],
            [
                "c.SampleConfigurable.trait_nohelp",
            ],
        ),
        (
            "autoclass/undoc_members.rst",
            [
                "c.SampleConfigurable.trait",
                "c.SampleConfigurable.trait_nohelp",
            ],
            [],
        ),
        (
            "autoconfigurable/exclude_members.rst",
            ["c.SampleConfigurable.trait_nohelp"],
            [
                "trait help text",
                "method docstring",
            ],
        ),
        (
            "autoconfigurable/inherited_members.rst",
            [
                "c.SampleConfigurableSubclass.trait",
                "c.SampleConfigurableSubclass.subclass_trait",
                "trait_noconfig",
                "method docstring",
            ],
            [],
        ),
        (
            "autoconfigurable/members.rst",
            [
                "c.SampleConfigurableSubclass.subclass_trait",
                "c.SampleConfigurableSubclass.trait_nohelp",
                "trait_noconfig",
                "No help string is provided.",
                "c.SampleConfigurableSubclass.trait",
                "method docstring",
            ],
            [
                "c.SampleConfigurableSubclass.trait_noconfig",
            ],
        ),
        (
            "autoconfigurable/no_members.rst",
            [
                "c.SampleConfigurableSubclass.subclass_trait",
                "c.SampleConfigurableSubclass.trait",
            ],
            [
                "trait_noconfig",
                "method docstring",
            ],
        ),
        ("autoconfigurable/non_configurable_raises_error.rst", [], []),
        (
            "autoconfigurable/specified_members.rst",
            [
                "method docstring",
                "c.SampleConfigurable.trait_nohelp",
                "trait help text",
            ],
            [],
        ),
        (
            "automodule/members.rst",
            [
                "sample_module docstring",
                "SampleConfigurable docstring",
                "SampleConfigurableSubclass docstring",
                "SampleNonConfigurable docstring",
                "SampleNonConfigurableSubclass docstring",
                "TraitTypesSampleConfigurable docstring",
            ],
            [],
        ),
        (
            "autotrait/help.rst",
            [
                "c.SampleConfigurable.trait",
                "trait help text",
                "trait_integer help text",
                "trait_unicode help text",
                "trait_dict help text",
                "trait_instance help text",
                "trait_union help text",
            ],
            [],
        ),
        (
            "autotrait/noconfig.rst",
            [
                "trait_noconfig",
                "Bool(False)",
            ],
            [
                "c.SampleConfigurable.trait_noconfig",
            ],
        ),
        (
            "autotrait/nohelp.rst",
            [
                "c.SampleConfigurable.trait_nohelp",
            ],
            [
                "No help string is provided.",
            ],
        ),
        ("autotrait/non_trait_raises_error.rst", [], []),
    ],
)
def test_sphinx_build_file(
    get_glob_filtered_temp_docs_dir,
    monkeypatch,
    rst_file_to_test,
    strings_in_html,
    strings_not_in_html,
):
    """
    Tests that individual .rst documents in the docs folder builds without
    warnings, emits .html with certain strings, and emits .html without certain
    strings.
    """
    temp_docs_dir = get_glob_filtered_temp_docs_dir(["index.rst", rst_file_to_test])
    monkeypatch.chdir(temp_docs_dir)

    p = subprocess.run(
        [
            "sphinx-build",
            "--color",
            "-W",
            "--keep-going",
            "-D",
            "exclude_patterns=",
            "source",
            "build",
        ],
        text=True,
    )
    if "raises_error" in rst_file_to_test:
        assert p.returncode > 0
        return
    assert p.returncode == 0

    html_file_to_inspect = os.path.join(
        "build", rst_file_to_test.replace(".rst", ".html")
    )
    with open(html_file_to_inspect) as f:
        html = f.read()

    for s in strings_in_html:
        assert s in html

    for s in strings_not_in_html:
        assert s not in html



import os
import sys

project = "autodoc_traits tests"
extensions = [
    "autodoc_traits",
    # sphinx.ext.napoleon is added to avoid warnings if testing with
    # :inherited-members:` where the super classe traitlets.HasTraits has numpy
    # or google-format docstrings that the napoleon can help interpret.
    "sphinx.ext.napoleon",
]

# ensure sample_module.py is on path
tests_dir = os.path.join(os.path.dirname(__file__), "..")
tests_dir = os.path.abspath(tests_dir)
sys.path.insert(0, tests_dir)

# Don't parse .rst documents expected to raise an error unless we want to test
# that specifically.
exclude_patterns = ["**/*_raises_error.rst"]
